"""Module for creation (and handling) of multidimensional vectors and trajectories."""
from __future__ import annotations

__author__ = "Matteo Gabba"
__copyright__ = "Copyright 2022, all right reserved Gabba Scientific"
__status__ = "Development"

from abc import abstractmethod
from collections.abc import Collection, Iterator
from dataclasses import dataclass
from pathlib import Path

import numpy as np

from gspy.decorators import check_args_len
from gspy.json_io import CustomJSONDict, JsonIOInterface
from gspy.time import Time, TimeList, TimeListLoader
from gspy.vectors import Vector, VectorListLoader, VectorsList

# TODO: create method to perform operations on time axis (invert, sort, rel_to_abs) in Trajectory
# TODO: create decorator to check type hinting of input args
# TODO: unit test the VectorNDim class


@dataclass(frozen=True)
class Trajectory(JsonIOInterface):
    """Represents a time series of 3D vectors instances.

    A Trajectory instance is instantiated by a list of 3D vectors and the corresponding times, a type error is raised
    otherwise. The two lists must contain the same number of Vector and Time instances.

    >>> source = Path.cwd() / "data" / "sub1.csv"
    >>> time_colnames = ['time']
    >>> coord_colnames = ['gaze_deg_coord_x', 'gaze_deg_coord_y']
    >>> times = TimeListLoader(source, usecols=time_colnames).load()
    >>> vectors = VectorListLoader(source, usecols=coord_colnames).load()
    >>> trajectory = Trajectory(vectors, times)
    >>> trajectory[:3]
    (VectorsList([Vector(0.4622481497169438, 3.76608427632374, 0.0),
                  Vector(0.46393828577858687, 3.7411490072433953, 0.0),
                  Vector(0.456283375837779, 3.7295640810489856, 0.0)]),
     TimeList([Time(1651585280.77981),
               Time(1651585280.78481),
               Time(1651585280.78981)]))

    A Trajectory instance contains the following properties: vector coordinates (x, y, z), vector magnitudes,
     time values, time zero (t0), relative time from start of the time-series:

    >>> trajectory.x[:5]
    array([0.46224815, 0.46393829, 0.45628338, 0.50202114, 0.51964137])
    >>> trajectory.y[:5]
    array([3.76608428, 3.74114901, 3.72956408, 3.70602158, 3.94218248])
    >>> trajectory.z[:5]
    array([0., 0., 0., 0., 0.])

    >>> trajectory.magnitude[:5]
    array([14.18500018, 13.99785081, 13.91120624, 13.73690808, 15.54314845])

    >>> trajectory.time[:5]
    array([1.65158528e+09, 1.65158528e+09, 1.65158528e+09, 1.65158528e+09,
           1.65158528e+09])

    >>> trajectory.t0
    1651585280.77981
    >>> trajectory.rtime[:5]
    array([0.        , 0.00500011, 0.00999999, 0.0150001 , 0.01999998])

    A Trajectory instance supports dot and cross vector products both trajectory-wise

    >>> dot_prod = trajectory.dot(trajectory)
    >>> dot_prod[:3]
    [14.397064128289646, 14.211434627409416, 14.117842753716685]

    >>> cross_prod = trajectory.cross(trajectory)
    >>> cross_prod[:3]
    (VectorsList([Vector(0.0, 0.0, 0.0),
                  Vector(0.0, 0.0, 0.0),
                  Vector(0.0, 0.0, 0.0)]),
     TimeList([Time(1651585280.77981),
               Time(1651585280.78481),

    and by a single vector

    >>> vector = trajectory[0][0]
    >>> vector
    Vector(0.4622481497169438, 3.76608427632374, 0.0)

    >>> dot_prod = trajectory.dot(vector)
    >>> dot_prod[:3]
    [14.397064128289646, 14.303937065747524, 14.256768789407998]
    >>> dot_prod = trajectory @ vector
    >>> dot_prod[:3]
    [14.397064128289646, 14.303937065747524, 14.256768789407998]

    >>> cross_prod = trajectory.cross(vector)
    >>> cross_prod[:3]
    (VectorsList([Vector(0.0, 0.0, 0.0),
                  Vector(0.0, 0.0, 0.01789147684168535),
                  Vector(0.0, 0.0, -0.005582448425092368)]),
     TimeList([Time(1651585280.77981),
               Time(1651585280.78481),
               Time(1651585280.78981)]))

    The following operations are implemented by operator overlaoding

    Get item(s) (Vector, Time) by indexing or slicing
    >>> trajectory[1]
    (Vector(0.46393828577858687, 3.7411490072433953, 0.0), Time(1651585280.78481))
    >>> trajectory[:2]
    (VectorsList([Vector(0.4622481497169438, 3.76608427632374, 0.0),
                  Vector(0.46393828577858687, 3.7411490072433953, 0.0)]),
    TimeList([Time(1651585280.77981), Time(1651585280.78481)]))

    Iteration
    >>> ((vector, time) for (vector, time) in trajectory)
    <generator object <genexpr> at 0x000002D749F9FB50>

    Get the total number of items
    >>> len(trajectory)
    6281

    Vector addition both trajectory-wise or by a single-vector
    >>> add1 = trajectory + trajectory
    >>> add1[:2]
    (VectorsList([Vector(0.9244962994338876, 7.53216855264748, 0.0),
                  Vector(0.9278765715571737, 7.4822980144867905, 0.0)]),
    TimeList([Time(1651585280.77981), Time(1651585280.78481)]))
    >>> add2 = trajectory + vector
    >>> add2[:2]
    (VectorsList([Vector(0.9244962994338876, 7.53216855264748, 0.0),
                  Vector(0.9261864354955307, 7.507233283567135, 0.0)]),
     TimeList([Time(1651585280.77981), Time(1651585280.78481)]))

    Vector subtraction both trajectory-wise or by a single-vector
    >>> sub1 = trajectory - trajectory
    >>> sub1[:2]
    (VectorsList([Vector(0.0, 0.0, 0.0), Vector(0.0, 0.0, 0.0)]),
     TimeList([Time(1651585280.77981), Time(1651585280.78481)]))
    >>> sub2 = trajectory - vector
    >>> sub2[:2]
    (VectorsList([Vector(0.0, 0.0, 0.0),
                  Vector(0.0016901360616430883, -0.024935269080344824, 0.0)]),
    TimeList([Time(1651585280.77981), Time(1651585280.78481)]))

    Coordinate-wise vector multiplication
    >>> mul1 = trajectory * trajectory
    >>> mul1[:2]
    (VectorsList([Vector(0.21367335191673806, 14.183390776372908, 0.0),
                  Vector(0.21523873301117374, 13.996195894398243, 0.0)]),
     TimeList([Time(1651585280.77981), Time(1651585280.78481)]))

    >>> mul2 = trajectory * vector
    >>> mul2[:2]
    (VectorsList([Vector(0.21367335191673806, 14.183390776372908, 0.0),
                  Vector(0.21445461418400247, 14.089482451563521, 0.0)]),
     TimeList([Time(1651585280.77981), Time(1651585280.78481)]))

    Vector scalar multiplication
    >>> mul3 = trajectory * 2
    >>> mul3[:2]
    (VectorsList([Vector(0.9244962994338876, 7.53216855264748, 0.0),
                  Vector(0.9278765715571737, 7.4822980144867905, 0.0)]),
     TimeList([Time(1651585280.77981), Time(1651585280.78481)]))

    Raise to the power of a number
    >>> pow1 = trajectory ** 2
    >>> pow1[:2]
    (VectorsList([Vector(0.21367335191673806, 14.183390776372908, 0.0),
                  Vector(0.21523873301117374, 13.996195894398243, 0.0)]),
     TimeList([Time(1651585280.77981), Time(1651585280.78481)]))

    Inverse vectors
    >>> inv = -trajectory
    >>> inv[:2]
    (VectorsList([Vector(-0.4622481497169438, -3.76608427632374, -0.0),
                  Vector(-0.46393828577858687, -3.7411490072433953, -0.0)]),
     TimeList([Time(1651585280.77981), Time(1651585280.78481)]))

    Absolute values
    >>> abs_val = abs(trajectory)
    >>> abs_val[:2]
    array([14.18500018, 13.99785081])

    Check (in)equality between all items of two Trajectory instances
    >>> trajectory == trajectory
    True
    >>> trajectory != trajectory
    False
    """

    vectors: VectorsList = VectorsList()
    times: TimeList = TimeList()

    @classmethod
    def from_json(cls, json_dict: CustomJSONDict) -> Trajectory:
        return cls(json_dict["vectors"], json_dict["times"])

    def __post_init__(self):
        if not isinstance(self.vectors, VectorsList):
            raise TypeError(
                f"{self.__class__.__name__}() argument 'vector' must be a type-checked list of Vector(s),"
                f" not '{self.vectors.__class__.__name__}'"
            )
        if not isinstance(self.times, TimeList):
            raise TypeError(
                f"{self.__class__.__name__}() argument 'times' must be a type-checked list of Time(s),"
                f" not '{self.times.__class__.__name__}'"
            )
        assert len(self.vectors) == len(self.times), (
            f"Trajectory instance arguments 'vectors' and 'times' must have same dimension, not {len(self.vectors)}"
            f" and {len(self.times)}"
        )

    @property
    def magnitude(self) -> np.ndarray:
        """Magnitudes of Trajectory vectors."""
        return np.array([vector.magnitude for vector in self.vectors])

    @property
    def x(self) -> np.ndarray:
        """X-coordinate of Trajectory vectors."""
        return np.array([vector.x for vector in self.vectors])

    @property
    def y(self) -> np.ndarray:
        """Y-coordinate of Trajectory vectors."""
        return np.array([vector.y for vector in self.vectors])

    @property
    def z(self) -> np.ndarray:
        """Z-coordinate of Trajectory vectors."""
        return np.array([vector.z for vector in self.vectors])

    @property
    def time(self) -> np.ndarray:
        return np.array([time.time for time in self.times])

    @property
    def rtime(self) -> np.ndarray:
        """Relative time from the first point of the time-series."""
        return self.time - self.t0

    @property
    def t0(self) -> float:
        """First point of the time series."""
        return self.time[0]

    @check_args_len
    def dot(self, other: Trajectory | Vector) -> list[float]:
        """Dot product of Trajectory vectors.

        If 'other' is a Trajectory instance the vectors of the two trajectories ('self' and 'other') are multiplied
        element by element.
        If 'other' is a Vector instance all vectors of 'self' are multiplied by the 'other' vector.

        Raise AssertionError if the two trajectories have different dimension.
        Raise TypeError if 'other' is not a Trajectory or Vector instance.

        Returns
        -------
            A list of float values corresponding to the dot products for each vector of the input Trajectory(s).
        """
        if isinstance(other, self.__class__):
            return [r1.dot(r2) for r1, r2 in zip(self.vectors, other.vectors)]
        if isinstance(other, Vector):
            return [r.dot(other) for r in self.vectors]
        raise TypeError(
            f"unsupported operand type(s) for {self.__class__.__name__}.dot():"
            f" '{other.__class__.__name__}'"
        )

    @check_args_len
    def cross(self, other: Trajectory | Vector) -> Trajectory:
        """Cross product of Trajectory vectors.

        If 'other' is a Trajectory instance the vectors of the two trajectories ('self' and 'other') are multiplied
        element by element.
        If 'other' is a Vector instance all vectors of 'self' are multiplied by the 'other' vector.

        Raise AssertionError if the two trajectories have different dimension.
        Raise TypeError if 'other' is not a Trajectory or Vector instance.

        Returns
        -------
            A Trajectory instance corresponding to the cross products for each vector of the input Trajectory(s).
        """
        if isinstance(other, self.__class__):
            return self(
                VectorsList(
                    [
                        r1.cross(r2)
                        for r1, r2 in zip(self.vectors, other.vectors)
                    ]
                ),
                self.times,
            )
        if isinstance(other, Vector):
            return self(
                VectorsList([r.cross(other) for r in self.vectors]),
                self.times,
            )
        raise TypeError(
            f"unsupported operand type(s) for {self.__class__.__name__}.cross():"
            f" '{other.__class__.__name__}'"
        )

    def __call__(self, vectors: VectorsList, times: TimeList) -> Trajectory:
        """Makes Trajectory instance callable."""
        return self.__class__(vectors, times)

    def __getitem__(self, index: int) -> tuple[Vector, Time]:
        """Get Trajectory items (Vector, Time) by index or slicing."""
        return self.vectors[index], self.times[index]

    def __iter__(self) -> Iterator[tuple[Vector, Time]]:
        """Allows iteration over the Trajectory items."""
        return iter(zip(self.vectors, self.times))

    def __len__(self) -> int:
        """Returns Trajectory length (number of items in the time-series)."""
        return len(self.vectors)

    @check_args_len
    def __add__(self, other: Trajectory | Vector) -> Trajectory:
        """Addition of Trajectory vectors.

        If 'other' is a Trajectory instance the vectors of the two trajectories ('self' and 'other') are added
        element by element.
        If 'other' is a Vector instance the 'other' vector is added to all vectors of 'self'.

        Raise AssertionError if the two trajectories have different dimension.
        Raise TypeError if 'other' is not a Trajectory or Vector instance.
        """
        if isinstance(other, self.__class__):
            return self(
                VectorsList(
                    [r1 + r2 for r1, r2 in zip(self.vectors, other.vectors)]
                ),
                self.times,
            )
        if isinstance(other, Vector):
            return self(
                VectorsList([r + other for r in self.vectors]), self.times
            )

        return NotImplemented

    def __radd__(self, other: Trajectory | Vector) -> Trajectory:
        return self + other

    @check_args_len
    def __sub__(self, other: Trajectory | Vector) -> Trajectory:
        """Subtraction of Trajectory vectors.

        If 'other' is a Trajectory instance the vectors of the two trajectories ('self' and 'other') are subtracted
        element by element.
        If 'other' is a Vector instance the 'other' vector is subtracted to all vectors of 'self'.

        Raise AssertionError if the two trajectories have different dimension.
        Raise TypeError if 'other' is not a Trajectory or Vector instance.
        """
        return self + (-other)

    def __rsub__(self, other: Trajectory | Vector) -> Trajectory:
        return other + (-self)

    @check_args_len
    def __mul__(
        self,
        other: Trajectory | Vector | int | float,
    ) -> Trajectory:
        """Coordinate-wise multiplication of Trajectory vectors.

        If 'other' is a Trajectory instance the vectors of the two trajectories ('self' and 'other') are multiplied
        coordinate-wise vector by vector.
        If 'other' is a Vector instance the 'other' vector is multiplied coordinate-wise to all vectors of 'self'.
        If 'other' is a number (float or int) all vectors of 'self' are scalar multiplied by 'other'.

        Raise AssertionError if the two trajectories have different dimension.
        Raise TypeError if 'other' is not a Trajectory, Vector, float or int type.
        """
        if isinstance(other, self.__class__):
            return self(
                VectorsList(
                    [r1 * r2 for r1, r2 in zip(self.vectors, other.vectors)]
                ),
                self.times,
            )
        if isinstance(other, (Vector, int, float)):
            return self(
                VectorsList([r * other for r in self.vectors]), self.times
            )
        return NotImplemented

    def __rmul__(self, other: Trajectory | int | float) -> Trajectory:
        return self * other

    def __matmul__(self, other: Trajectory | Vector) -> list[float]:
        """Overload @ operator with dot product."""
        return self.dot(other)

    def __pow__(self, power: int | float) -> Trajectory:
        """Raise to the power Trajectory vectors."""
        if isinstance(power, (int, float)):
            return self(
                VectorsList([r**power for r in self.vectors]), self.times
            )
        return NotImplemented

    def __neg__(self) -> Trajectory:
        """Inverse of Trajectory vectors."""
        return self(VectorsList([-r for r in self.vectors]), self.times)

    def __pos__(self) -> Trajectory:
        return self

    def __abs__(self) -> np.ndarray:
        """Absolute value (magnitude) of Trajectory vectors."""
        return self.magnitude

    @check_args_len
    def __eq__(self, other: Trajectory) -> bool:
        """Check equality between each item (Vector, Time) of two Trajectory instances."""
        if isinstance(other, self.__class__):
            return (self.vectors == other.vectors) and (
                self.times == other.times
            )
        return NotImplemented

    def __ne__(self, other: Trajectory) -> bool:
        """Check inequality between each item (Vector, Time) of two Trajectory instances."""
        return not self == other

    def __hash__(self) -> int:
        """Return hash value of the given instance."""
        return hash(tuple(self))

    def __bool__(self) -> bool:
        """Check that all vectors and times are null."""
        return all(r == 0 for r, t in self) and all(t == 0 for r, t in self)

    def __repr__(self) -> str:
        """Returns a representation of a Vector object."""
        return f"{self.__class__.__name__}({[(r, t) for r, t in zip(self.vectors, self.times)]})"

    def __str__(self) -> str:
        """Returns a string representation of a Vector object."""
        return f"length={len(self)}, \n{[(r, t) for r, t in zip(self.vectors, self.times)]}"


class TrajectoryLoader:
    """Create a Trajectory instance from file.

    >>> source = Path.cwd() / "data" / "sub1.csv"
    >>> time_colnames = ['time']
    >>> coord_colnames = ['gaze_deg_coord_x', 'gaze_deg_coord_y']
    >>> trajectory_loader = TrajectoryLoader(source, time_colnames, coord_colnames)
    >>> trajectory = trajectory_loader.load()
    >>> trajectory[:2]
    (VectorsList([Vector(0.4622481497169438, 3.76608427632374, 0.0),
                  Vector(0.46393828577858687, 3.7411490072433953, 0.0)]),
     TimeList([Time(1651585280.77981), Time(1651585280.78481)]))
    """

    def __init__(
        self,
        source: Path,
        time_colnames: list[str],
        coord_colnames: list[str],
    ) -> None:
        """
        ---

            source : Path, required
                Path of the file storing the vector coordinates and the corresponding times.

            time_colnames : list[str], required
                Name of the column storing the time values.

            coord_colnames : list[str], required
                Names of the columns storing the 3D coordinates, or a subset of those, of the vectors.
        """
        self.source = source
        self.time_colnames = time_colnames
        self.coord_colnames = coord_colnames

    def load(self) -> Trajectory:
        """
        ---

            A Trajectory instance.
        """
        times = TimeListLoader(self.source, usecols=self.time_colnames).load()
        vectors = VectorListLoader(
            self.source, usecols=self.coord_colnames
        ).load()

        return Trajectory(vectors, times)


@dataclass
class BaseTrajectoryCollection(Collection):
    def __post_init__(self):
        for name, value in self.__dict__.items():
            if not isinstance(value, Trajectory):
                raise TypeError(
                    f"unsupported '{name}' argument type for {self.__class__.__name__}():"
                    f" '{value.__class__.__name__}'"
                )

    def __iter__(self) -> Iterator[tuple[str, Trajectory]]:
        for (name, trajectory) in self.__dict__.items():
            yield name, trajectory

    def __contains__(self, trajectory) -> bool:
        return trajectory in self.__dict__.values()

    def __len__(self) -> int:
        return len(self.__dict__)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(attrs={[key for key in self.__dict__.keys()]})"



"""Module with general utility functions."""
from __future__ import annotations

__author__ = "Matteo Gabba"
__copyright__ = "Copyright 2022, all right reserved Gabba Scientific"
__status__ = "Development"

import itertools
import math
from collections.abc import Iterator
from typing import Iterable

import more_itertools
import numpy as np


def bin_edge_to_center(bin_edges: np.ndarray) -> np.ndarray:
    """
    Convert bin edges to bin center.

    Parameters
    ----------
        bin_edges : 'array'
            Bin edges. For example as defined by numpy.histogram().

    Returns
    -------
        bin_center : 'array'
            Bin center position calculated as (left_edges + right_edges) / 2.
    """
    # Compute the bin center:
    left_edges = bin_edges[1:]
    right_edges = bin_edges[:-1]
    bin_center = (left_edges + right_edges) / 2

    return bin_center


def build_histogram(
    data: np.ndarray,
    hist_range: tuple[float, float],
    bins: int | str = "auto",
    density: bool = True,
) -> tuple[np.ndarray, np.ndarray, float]:

    values, bin_edges = np.histogram(
        data,
        range=hist_range,
        bins=bins,
        density=density,
    )

    bin_centers = bin_edge_to_center(bin_edges)
    bin_width = bin_edges[1] - bin_edges[0]

    return values, bin_centers, bin_width


def cluster_size_generator(
    values: Iterable[float],
) -> Iterator[tuple[float, np.ndarray, np.ndarray]]:
    """
    Yield generator with value, counts and index of groups of consecutive equal values in a 1D-list.

    Adapted from:
        https://stackoverflow.com/questions/23023805/how-to-get-the-index-and-occurance-of-each-item-using-itertools-groupby

    Parameters
    ----------
        values : list
            1D list of values.

    Returns
    -------
        (value, counts, idx) : generator
    """
    # Set index indicating start of the first group
    group_start_idx = 0

    # For each group of consecutive equal values
    for value, group in itertools.groupby(values):
        # counts the consecutive equal values
        counts = len(list(group))

        # generate array of index corresponding to each group of consecutive values
        idx = np.arange(group_start_idx, group_start_idx + counts)

        # yield generator with value, counts and index of each group
        yield value, counts, idx

        # re-set the start index for next group
        group_start_idx += counts


def _isnan(value: int | float) -> bool:
    if math.isnan(value):
        return True
    return False


def pop_iterator(iterator: Iterator) -> Iterator:
    """Pop last item from iterator object."""
    iter_1, iter_2 = itertools.tee(iterator, 2)
    n_items = more_itertools.ilen(iter_1)
    return more_itertools.take(n_items - 1, iter_2)



"""Module
"""
from __future__ import annotations

__author__ = "Matteo Gabba"
__copyright__ = "Copyright 2022, all right reserved Gabba Scientific"
__status__ = "Development"

from functools import wraps


def check_args_len(method):
    """Class-method decorator to assert length of two vector arguments (self, other)."""

    @wraps(method)
    def wrapped_method(self, other):
        method_name = method.__qualname__

        if isinstance(other, self.__class__):
            assert len(self) == len(other), (
                f"{method_name}(self, other): operand 'self' has different dimension"
                f" ({len(self)}) than operand 'other' ({len(other)})"
            )

        return method(self, other)

    return wrapped_method



"""Module for signal processing."""
from __future__ import annotations

__author__ = "Matteo Gabba"
__copyright__ = "Copyright 2022, all right reserved Gabba Scientific"
__status__ = "Development"

import math
import warnings
from dataclasses import dataclass
from typing import Iterable

import numpy as np
from pywt import wavedec, waverec
from scipy import signal
from scipy.integrate import cumulative_trapezoid
from statsmodels.tsa.ar_model import ar_select_order
from statsmodels.tsa.stattools import adfuller

from gspy.general import TypeCheckedList
from gspy.utils import _isnan, cluster_size_generator


@dataclass
class Peak:
    is_type: str = "unknown"
    index: int = None
    position: float = None
    fwhm: float = None
    std: float = None
    width_height: float = None
    width_start: float = None
    width_end: float = None
    height: float = None
    contour_height: float = None
    prominence: float = None

    @property
    def area(self) -> float:
        """See: # https://www.savarese.org/math/gaussianintegral.html#S3.E20."""
        return self.prominence * self.std * np.sqrt(2 * math.pi)

    def __lt__(self, other: Peak) -> bool:
        if isinstance(other, Peak):
            return self.position < other.position
        return NotImplemented

    def __eq__(self, other: Peak) -> bool:
        if isinstance(other, Peak):
            return (
                (self.index == other.index)
                & (self.position == other.position)
                & (self.fwhm == other.fwhm)
                & (self.height == other.height)
            )
        return NotImplemented

    def __ne__(self, other: Peak) -> bool:
        return not self == other

    def __hash__(self):
        return hash((self.is_type, self.position, self.std, self.height))

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(type={self.is_type}, position={self.position:.1f})"


class PeaksList(TypeCheckedList):
    def __init__(
        self,
        iterator_arg: Iterable[Peak] = None,
    ):
        """
        Iterator_arg : List[Peak], optional, default to None.

            List of Peak
        """
        # Initialize the parent class to type check for Peak instances
        super().__init__(iterator_arg=iterator_arg, instance_type=Peak)

    @property
    def indexes(self) -> np.ndarray:
        return np.array([peak.index for peak in self])

    @property
    def positions(self) -> np.ndarray:
        return np.array([peak.position for peak in self])

    @property
    def fwhms(self) -> np.ndarray:
        return np.array([peak.fwhm for peak in self])

    @property
    def stds(self) -> np.ndarray:
        return np.array([peak.std for peak in self])

    @property
    def width_heights(self) -> np.ndarray:
        return np.array([peak.width_height for peak in self])

    @property
    def width_starts(self) -> np.ndarray:
        return np.array([peak.width_start for peak in self])

    @property
    def width_ends(self) -> np.ndarray:
        return np.array([peak.width_end for peak in self])

    @property
    def heights(self) -> np.ndarray:
        return np.array([peak.height for peak in self])

    @property
    def prominences(self) -> np.ndarray:
        return np.array([peak.prominence for peak in self])

    @property
    def contour_heights(self) -> np.ndarray:
        return np.array([peak.contour_height for peak in self])

    @property
    def areas(self) -> np.ndarray:
        return np.array([peak.area for peak in self])

    def compute_features(
        self,
        y: np.ndarray,
        x: np.ndarray,
    ) -> None:
        compute_peaks_width_features(
            y,
            x,
            self,
        )

        compute_peaks_height_features(
            y,
            self,
        )

    def detect(
        self,
        y: np.ndarray,
        x: np.ndarray,
        min_peak_prominence: float,
        min_inter_peak_distance: float,
    ) -> None:
        """Populate input list with detected peaks."""
        inter_sample_distance = x[1] - x[0]
        min_inter_peak_distance_cts = int(
            min_inter_peak_distance / inter_sample_distance
        )

        peaks_indexes, _ = signal.find_peaks(
            y,
            prominence=min_peak_prominence,
            distance=min_inter_peak_distance_cts,
        )
        peaks_iter = (
            Peak(index=index, position=x[index]) for index in peaks_indexes
        )

        self.extend(peaks_iter)


def compute_peaks_width_features(
    y: np.ndarray,
    x: np.ndarray,
    peaks: PeaksList,
) -> None:
    inter_sample_distance = x[1] - x[0]
    x0 = x[0]

    results_fwhm = signal.peak_widths(
        y,
        peaks.indexes,
        rel_height=0.5,
    )
    fwhms = results_fwhm[0] * inter_sample_distance
    stds = fwhms / 2.355
    width_heights = results_fwhm[1]
    width_starts = (results_fwhm[2] * inter_sample_distance) + x0
    width_ends = width_starts + fwhms

    for n, peak in enumerate(peaks):
        peak.fwhm = fwhms[n]
        peak.std = stds[n]
        peak.width_height = width_heights[n]
        peak.width_start = width_starts[n]
        peak.width_end = width_ends[n]


def compute_peaks_height_features(
    y: np.ndarray,
    peaks: PeaksList,
) -> None:
    prominences = signal.peak_prominences(y, peaks.indexes)[0]
    contour_heights = y[peaks.indexes] - prominences
    heights = y[peaks.indexes]

    for n, peak in enumerate(peaks):
        peak.prominence = prominences[n]
        peak.contour_height = contour_heights[n]
        peak.height = heights[n]


def get_weights(n_train: int, n_steps: int) -> np.ndarray:
    """
    Compute weights (reliability) of predicted values.

    Parameters
    ----------
        n_train : int
            Number of points used for training the AR-model and making predictions.

        n_steps : int
            Number of predicted points.

    Returns
    -------
        w : np.ndarray
            Weights indicating the reliability of the predicted values, dim(n_steps,).
            The weights are linearly proportional to the number of points used for the prediction
            and inversely proportional to the number of steps into the future.
    """
    # Total weights: are linearly proportional to the number of points used for the prediction
    # and inversely proportional to the number of steps into the future
    linear_w = np.arange(1, n_steps + 1)[::-1]

    return n_train * linear_w


def autoregress_predict(
    seg: np.ndarray, n_steps: int, max_lag: int
) -> np.ndarray:
    """
    Predict (forecast) n-steps into the future using an autoregressive (AR) model that minimizes the Akaike information.

     criterion (AIC).
     # TODO: implement possibility to choose the width of the region for prediction.

    See:
        https://www.statsmodels.org/devel/examples/notebooks/generated/autoregressions.html
        https://vitalflux.com/autoregressive-ar-models-with-python-examples/#:~:text=Autoregressive%20(AR)%20models%20are%20a,order%20to%20make%20accurate%20predictions.
        https://machinelearningmastery.com/autoregression-models-time-series-forecasting-python/

    Parameters
    ----------
        seg : np.ndarray
            Segment of a time-series used for training the AR-model.

        n_steps : np.ndarray
            Forecast n-steps into the future based on the past behavior of the time-series.

        max_lag : int
            Select max lag for automatic selection of the lags structure.
            See: https://www.statsmodels.org/devel/generated/statsmodels.tsa.ar_model.ar_select_order.html

    Returns
    -------
        prediction : np.ndarray
            Array of len = n_steps with AR-predicted values.
    """
    # Number of points used for training of the AR-model
    n_samples = len(seg)

    # Check for stationarity of the time-series data
    # We will look for p-value. In case, p-value is less than 0.05, the time series
    # data can be said to have stationarity
    df_stationarityTest = adfuller(seg, autolag="AIC")
    p_value = df_stationarityTest[1]

    # Raise warning if time-series segment used for prediction is not stationary
    if p_value > 0.05:
        warnings.warn(
            "P-value = {} is > 0.05! The time-series segment used for prediction may "
            "not be stationary!".format(p_value)
        )

    # Automatically select AR-model lag
    select = ar_select_order(seg, ic="aic", maxlag=max_lag)

    # Train AR-model
    trained_model = select.model.fit()

    # Make predictions (forecast) to refill gap
    prediction = trained_model.predict(
        start=n_samples, end=n_samples + n_steps - 1, dynamic=True
    )

    return prediction


def find_gaps_and_segments(x: np.ndarray | list[float]) -> list[np.ndarray]:
    """
    Get indexes of gaps (of NaN's) and segments (not-NaN) in the input array.

    Parameters
    ----------
        x : array-like
            Array with NaN gaps.

    Returns
    -------
        A list of indexes arrays corresponding to the segments and gaps in the input array.
    """
    # Get indexes of segments (not-NaN) and gaps (NaN) in the array
    return [idx for (value, _, idx) in cluster_size_generator(np.isnan(x))]


def gap_is_first(x: np.ndarray | list[float]) -> bool:
    """
    Check if a gap (of NaN's) or a segment is in first position of the input array.

    Parameters
    ----------
        x : array-like
            Array with NaN gaps.

    Returns
    -------
        True if the input array starts with a gap of NaN's, False, otherwise.
    """
    # If a gap is in first position: gap-segment-gap-segment-etc..
    if _isnan(x[0]):
        return True

    # if a segment is in first position: segment-gap-segment-etc..
    return False


def autoregress_fill(
    x: np.ndarray, max_lag: int
) -> tuple[np.ndarray, np.ndarray]:
    """
    Fill gaps of NaN's in a time-series using an autoregressive model that minimizes the Akaike information.

    criterion (AIC).
    # TODO: implement possibility to choose the width of the region for prediction.

    Parameters
    ----------
        x : array-like
            Array with NaN gaps, dim(n,).

        max_lag : int
            Select max lag for automatic selection of the lags structure.

    Returns
    -------
        y : np.ndarray
            A copy of the input array with gaps filled with the AR-predicted values, dim(n,).

        w : np.ndarray
            Weights indicating the reliability of the predicted values, dim(n,).
    """
    # Copy input array
    y = np.copy(x)

    # Initialize weights
    w = np.ones_like(x)

    # Get labels and indexes of NaN gaps (label: True) and segments (label: False)
    # in the input array
    idx = find_gaps_and_segments(y)

    # Total number of segments and gaps
    n = len(idx)

    # Initialize variables
    idx_first_seg = 0
    prediction = np.empty(0)
    w_i = np.empty(0)

    # Array starts with a gap
    if gap_is_first(y):

        # position of the first segment in the array
        idx_first_seg = 1

        # indexes of first gap
        idx_first_gap = idx[0]
        first_gap = y[idx_first_gap]

        # nothing to predict
        prediction = np.zeros_like(first_gap)

        # fill first gap with zeros
        y[idx_first_gap] = prediction
        w[idx_first_gap] = np.zeros_like(first_gap)

    # array starts with a segment
    elif not gap_is_first(y):

        # position of the first segment in the array
        idx_first_seg = 0

    # Predict from segment immediately prior to the gap starting from first segment in the array
    for i in range(idx_first_seg, n - 1, 2):

        # get segment and gap indexes
        idx_seg = idx[i]
        idx_gap = idx[i + 1]

        # initialize segment and gap arrays with right length
        seg = y[idx_seg]
        gap = y[idx_gap]

        # for a segment only one sample long
        if len(seg) == 1:

            # predict a constant
            prediction = np.ones_like(gap) * seg[0]

        # for a segment longer than one sample
        elif len(seg) > 1:

            # predict with autoregressive model
            prediction = autoregress_predict(seg, len(gap), max_lag)
            w_i = get_weights(len(seg), len(gap))

        # fill gap with predicted values
        y[idx_gap] = prediction
        w[idx_gap] = w_i

    return y, w


def refill_gaps(x: np.ndarray, max_lag: int) -> np.ndarray:
    """
    Replaces NaN values of an input vector, X, by fitting an autoregressive (AR) model that.

    minimizes the Akaike information criterion. Each NaN value is replaced by a weighted average
    of the values estimated by forward and backward prediction.

    # TODO: implement possibility to choose the width of the region for prediction.

    Parameters
    ----------
        x : array-like
            Array with NaN gaps, dim(n,).

        max_lag : int
            Select max lag for automatic selection of the lags structure.

    Returns
    -------
        y : np.ndarray
            A copy of the input array with NaN gaps refilled with AR-predicted values.
    """
    # Refill NaN's gaps in forward direction
    x_fwd, w_fwd = autoregress_fill(x, max_lag)

    # Refill NaN's gaps in backward direction
    x_bwd, w_bwd = autoregress_fill(x[::-1], max_lag)

    # Invert refilled array
    x_bwd = x_bwd[::-1]
    w_bwd = w_bwd[::-1]

    # Compute weighted average of the refilled arrays in backward and forward directions
    y = (x_bwd * w_bwd + x_fwd * w_fwd) / (w_bwd + w_fwd)

    return y


def cumulative_integral(
    y: np.ndarray,
    x: np.ndarray = None,
    dx: float = None,
    max_lag: int = 5,
) -> np.ndarray:
    y_filled = refill_gaps(y, max_lag)
    return cumulative_trapezoid(y_filled, x, dx, initial=0)


def central_diff(
    y: np.ndarray | list[float],
    axis: int = -1,
    kernel_half_size: int = 1,
) -> np.ndarray:
    """
    Compute sum of central finite differences of input array along the given axis using a convolution kernel.

    The half width M of the kernel determines the number of terms in the summation:
    dy[i] = sum_m (dy[i+m] - dy[i-m]) with m in [-M, M].

    See: https://numpy.org/doc/stable/reference/generated/numpy.convolve.html

    Keep in mind that: "the convolution operator flips the second array before “sliding” the two across one another."

    Parameters
    ----------
        y : array-like
            One or multidimensional array.

        axis : int
            The axis along which the difference is taken, default is the last axis.

        kernel_half_size : int
            Half size M of the family of convolution kernels: g[m] = [1, 0, -1], [1, 1, 0, -1, -1], ...

    Returns
    -------
        dy : np.ndarray
        The summation of the m-th central finite differences of the input array:
        dy[i] = sum_m (dy[i+m] - dy[i-m]) with m in [-M, M]
        The output array is padded with NaN along the chosen axis to keep the same dimensions of the input array.

    Version history
    ---------------
        v1 : 18-05-2022 (m.gabba)
    """
    assert (
        kernel_half_size > 0
    ), "kernel_half_size: must be a not-zero positive integer!"

    # Base convolution kernel
    base_kernel = [0]

    # Create convolution kernel for calculation of central difference with chosen half-width histrange: M // 2
    kernel = np.pad(
        base_kernel,
        pad_width=kernel_half_size,
        mode="constant",
        constant_values=[1, -1],
    )

    # Apply convolution kernel along the given axis
    dy = np.apply_along_axis(
        lambda m: np.convolve(m, kernel, mode="valid"), axis=axis, arr=y
    )

    # Pad with NaN along the given axis to keep the same dimensions of the input array.
    dy_padded = np.apply_along_axis(
        lambda m: np.pad(
            m,
            pad_width=kernel_half_size,
            mode="constant",
            constant_values=np.nan,
        ),
        axis=axis,
        arr=dy,
    )

    return dy_padded


def smooth_signal_by_differentiation_and_reintegration(
    y: np.ndarray,
    x: np.ndarray,
    kernel_half_size: int,
) -> np.ndarray:
    delta_y = central_diff(y, kernel_half_size=kernel_half_size)
    delta_x = central_diff(x, kernel_half_size=kernel_half_size)
    derivative = delta_y / delta_x
    return cumulative_integral(derivative, x)


def correlation(
    sig1: np.ndarray | list[float],
    sig2: np.ndarray | list[float],
    freq: float,
    normed: bool,
) -> tuple[np.ndarray, np.ndarray, float]:
    """
    Compute correlation between two signals.

    See:
        https://en.wikipedia.org/wiki/Cross-correlation
        https://scicoding.com/cross-correlation-in-python-3-popular-packages/
        https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.correlate.html

    Parameters
    ----------
        sig1, sig2 : array-like
            Signal(s) to correlate. If sig1=sig2 returns the auto-correlation; otherwise,
            the cross-correlation.

        freq : float
            Sampling frequency of the input signal(s).

        normed : bool
            If True return the normalized correlation with values between [-1, 1]
            as per definition; if False, the not-normalized correlation.

    Returns
    -------
        corr, lag_times, lags, lag : tuple
            The correlation function, the lag times and the lag time corresponding to the max
            correlation.
    """
    # Length of the signals
    n1 = len(sig1)
    n2 = len(sig2)

    # Length of longer signal
    N = np.max([n1, n2])

    if normed:

        # subtract mean signal value
        nsig1 = sig1 - np.mean(sig1)
        nsig2 = sig2 - np.mean(sig2)

        # correlate normed signals
        corr = signal.correlate(nsig1, nsig2, mode="full")

        # normalize correlation between [-1, 1] using the signal std
        norm = N * np.std(nsig1) * np.std(nsig2)
        corr /= norm

    if not normed:

        # correlate raw signals
        corr = signal.correlate(sig1, sig2, mode="full")

        # divide by the number of sample
        corr /= N

    # Get lag indixes: k
    lags = signal.correlation_lags(n1, n2)

    # Convert lag indexes to time: k * dt = k / freq
    lag_times = lags / freq

    # Lag time corresponding to the max correlation
    lag = lag_times[np.argmax(corr)]

    return corr, lag_times, lag


def wavelet_filter(
    signal: np.ndarray,
    wavelet_type: str,
    extension_mode: str,
    filtered_levels: tuple[int, ...],
) -> tuple[np.ndarray, list[float]]:
    """
    Filter signal with discrete wavelet transformation (DWT) and reconstruction.

    See : https://ataspinar.com/2018/12/21/a-guide-for-using-the-wavelet-transform-in-machine-learning/

    Parameters
    ----------
        signal : np.ndarray
            Input data series.

        wavelet_type : wavelet object or str
            Valid discrete wavelet type.
                See :
                    https://pywavelets.readthedocs.io/en/latest/ref/wavelets.html
                    http://wavelets.pybytes.com/wavelet/

        extension_mode : str
            Signal extension mode.
                See : https://pywavelets.readthedocs.io/en/latest/ref/signal-extension-modes.html#ref-modes

        filtered_levels : tuple
            Choose decomposition levels to remove from signal reconstruction in order to filter specific frequencies.
            Levels are positive integers. Low values corresponds to high-frequency wavelets, high values to
             low-frequency wavelets. Thus, removing low (high) value levels corresponds to a  low(high)pass wavelet
              filter.
                See : https://pywavelets.readthedocs.io/en/latest/ref/idwt-inverse-discrete-wavelet-transform.html

    Returns
    -------
        reconstructed_signal : np.ndarray
            The input signal after wavelet decomposition and recomposition with filtered coefficients.

         coefficients : List
            The wavelet decomposition coefficients after filtering.

    Version history
    ---------------
        v1 : 18-05-2022 (m.gabba)
    """
    # Discrete wavelet transformation DWT of signal with automatic selection of the decomposition level
    # See: https://pywavelets.readthedocs.io/en/latest/ref/dwt-discrete-wavelet-transform.html
    # The output coefficients are ordered as follows: [cA_n, cD_n, cD_n-1, …, cD2, cD1] with decreasing level order
    coefficients = wavedec(
        signal, wavelet_type, mode=extension_mode, level=None
    )

    # Set to zero coefficients at the chosen levels
    for level in filtered_levels:
        coefficients[-level] = np.zeros_like(coefficients[-level])

    # Reconstruct wavelet filtered signal
    reconstructed_signal = waverec(
        coefficients, wavelet_type, mode=extension_mode
    )

    # Handle issues with reconstructed signal longer than the original signal, I don't understand the origin of
    # this behavior!!

    # If the reconstructed signal is longer than the original signal
    if len(reconstructed_signal) > len(signal):
        # remove last point
        reconstructed_signal = reconstructed_signal[:-1]

    return reconstructed_signal, coefficients



"""Module for handling data input/output from and to files."""
from __future__ import annotations

__author__ = "Matteo Gabba"
__copyright__ = "Copyright 2022, all right reserved Gabba Scientific"
__status__ = "Development"

import json
from abc import ABC, abstractmethod
from collections.abc import Iterable, Mapping
from pathlib import Path
from pydoc import locate
from typing import Any

import numpy as np

JSONType = str, dict | int | float | str | list | bool | None


class CustomJSONDict(dict):
    """Dictionary satisfying JSON decodable / encodable interface."""

    def __init__(
        self,
        data: Mapping[str, JSONType] | Iterable[str, JSONType] = None,
        **kwargs: JSONType,
    ) -> None:
        if not data:
            super().__init__()
        if data:
            super().__init__(data)
        if kwargs:
            super().__init__(**kwargs)
        self["__class__"] = None


class JsonIOInterfaceError(BaseException):
    pass


class CustomJSONEncoder(json.JSONEncoder):
    def default(self, obj: Any) -> CustomJSONDict | JSONType:

        if isinstance(obj, np.integer):
            return int(obj)

        if isinstance(obj, type):
            return str(obj.__name__)

        if isinstance(obj, Path):
            json_dict = CustomJSONDict(path=str(obj))
            json_dict["__class__"] = "pathlib.Path"
            return json_dict

        if isinstance(obj, np.ndarray):
            json_dict = CustomJSONDict(array=obj.tolist())
            json_dict["__class__"] = "numpy.ndarray"
            return json_dict

        else:
            try:
                return obj.to_json()
            except AttributeError as error:
                raise JsonIOInterfaceError(
                    f"missing method to_json() in class: '{obj.__class__.__name__}'"
                ) from error
            except TypeError:
                pass

        # Let the base class default method raise the TypeError
        return super().default(obj)


class CustomObjectJSONDumper:
    """Dump custom objects to JSON file."""

    def __init__(self, output_path: Path, obj: Any) -> None:
        self.output_path = output_path
        self.obj = obj

    def dump(self) -> None:
        with open(self.output_path, "w", encoding="utf-8") as file:
            json_data = json.dumps(self.obj, cls=CustomJSONEncoder, indent=4)
            file.write(json_data)


class CustomJSONDecoder(json.JSONDecoder):
    def __init__(self, **kwargs):
        kwargs["object_hook"] = self.object_hook
        super().__init__(**kwargs)

    def object_hook(self, json_dict: CustomJSONDict) -> Any:

        try:
            __class__ = json_dict.pop("__class__")
        except KeyError as error:
            raise JsonIOInterfaceError(
                f"missing field name '__class__' in JSON object: {json_dict}"
            ) from error

        try:
            cls = locate(__class__)
        except KeyError as error:
            raise JsonIOInterfaceError(
                f"JSON object '{__class__}' not found"
            ) from error

        # Handle patlib.Path
        if cls is Path:
            obj = Path(json_dict["path"])
            return obj

        # Handle numpy.ndarray
        if cls is np.ndarray:
            obj = np.array(json_dict["array"])
            return obj

        if hasattr(cls, "from_json"):
            obj = cls().from_json(json_dict)
            return obj

        else:
            raise JsonIOInterfaceError(
                f"missing method from_json() in class: '{cls.__class__.__name__}'"
            )


class JSONLoader:
    """Load JSON object using CustomJSONEncoder."""

    def __init__(self, input_path: Path) -> None:
        self.input_path = input_path

    @staticmethod
    def _parse_invalid_json_values(value: str) -> float:
        conversion = {
            "-Infinity": -float("inf"),
            "Infinity": float("inf"),
            "NaN": float("nan"),
        }
        return conversion[value]

    def load(self) -> Any:
        with open(self.input_path, "r", encoding="utf-8") as file:
            json_data = file.read()
            obj = json.loads(
                json_data,
                cls=CustomJSONDecoder,
                parse_constant=self._parse_invalid_json_values,
            )

        return obj


class BaseJsonIOInterface(ABC):
    """Interface of JSON encodable / decodable objects."""

    @classmethod
    @abstractmethod
    def from_json(cls, json_dict: CustomJSONDict) -> Any:
        ...

    @abstractmethod
    def to_json(self) -> CustomJSONDict:
        ...


class DefaultJsonIOMixin:
    @classmethod
    def from_json(cls, json_dict: CustomJSONDict) -> Any:

        obj = cls()

        for key, value in json_dict.items():

            try:
                getattr(obj, key)
            except AttributeError as error:
                raise AttributeError(
                    f"invalid attribute '{key}' for class: '{cls.__class__.__name__}'"
                ) from error

            try:
                __class__ = value.pop("__class__")
            except (KeyError, AttributeError, TypeError):
                obj.__dict__[key] = value
                continue

            try:
                _cls = locate(__class__)
            except KeyError as error:
                raise JsonIOInterfaceError(
                    f"class '{__class__}' not found'"
                ) from error

            # Handle patlib.Path
            if _cls is Path:
                obj.__dict__[key] = Path(value)
                continue

            # Handle numpy.ndarray
            if _cls is np.ndarray:
                obj.__dict__[key] = np.array(value)
                continue

            # Handle custom objects
            if hasattr(_cls, "from_json"):
                try:
                    obj.__dict__[key] = json.loads(
                        value, cls=CustomJSONDecoder
                    )
                except TypeError:
                    pass

            elif not hasattr(_cls, "from_json"):
                raise JsonIOInterfaceError(
                    f"missing method from_json() in class: '{obj.__class__.__name__}'"
                )

        # Let the base class default method raise the TypeError
        return obj

    def _get__class__(self) -> str:
        return ".".join([self.__module__, self.__class__.__name__])

    def to_json(self) -> CustomJSONDict:
        json_dict = CustomJSONDict(self.__dict__)
        json_dict["__class__"] = self._get__class__()
        return json_dict


class JsonIOInterface(DefaultJsonIOMixin, BaseJsonIOInterface):
    pass



from __future__ import annotations

from collections import UserList
from collections.abc import Iterable
from typing import Any

from gspy.json_io import CustomJSONDict, JsonIOInterface

# TODO: add docstrings and type annotations


class TypeCheckedList(UserList, JsonIOInterface):
    """
    Type checked list of instances of the same class.

    See: https://stackoverflow.com/questions/3487434/overriding-append-method-after-inheriting-from-a-python-list
    """

    def __init__(
        self, iterator_arg: Iterable[Any] = None, instance_type: Any = float
    ) -> None:

        super().__init__()
        self._instance_type = instance_type
        if iterator_arg is not None:
            self.extend(iterator_arg)  # This validates the arguments...

    def to_json(self, obj_dict=None) -> CustomJSONDict:
        json_dict = CustomJSONDict({"data": list(self)})
        json_dict["__class__"] = self._get__class__()
        return json_dict

    def _check_type(self, instance):
        if not isinstance(instance, self._instance_type):
            raise TypeError(
                f"{self.__class__.__name__} is a collection of {self._instance_type} instances"
                f" but {instance} ({instance.__class__}) was passed instead"
            )

        return instance

    def insert(self, i, item):
        return super().insert(i, self._check_type(item))

    def append(self, item):
        return super().append(self._check_type(item))

    def extend(self, other):
        return super().extend([self._check_type(v) for v in other])

    def __call__(self, iterator_arg: Iterable[Any]) -> TypeCheckedList:
        return self.__class__(iterator_arg)

    def __setitem__(self, i, other):
        if isinstance(i, slice):
            return super().__setitem__(
                i, [self._check_type(v) for v in other]
            )  # Extended slice...
        return super().__setitem__(i, self._check_type(other))

    def __add__(self, other):
        return super().__add__([self._check_type(v) for v in other])

    def __iadd__(self, other):
        return super().__iadd__([self._check_type(v) for v in other])

    def __repr__(self):
        return f"{self.__class__.__name__}({[item for item in self]})"
